import os
import tarfile
import base64
import json
from time import sleep
from subprocess import call
from paver.easy import *
from paver.setuputils import setup, find_packages

try:
    from http import client as http
    from urllib.request import urlretrieve
except ImportError:
    from urllib import urlretrieve
    import httplib as http

setup(
    name='neo4jdb',
    version='0.0.9',
    author='Jacob Hansson',
    author_email='jakewins@gmail.com',
    packages=find_packages(),
    py_modules=['setup'],
    include_package_data=True,
    install_requires=[],
    url='https://github.com/jakewins/neo4jdb-python',
    description='DB API 2.0 driver for the Neo4j graph database.',
    long_description=open('README.rst').read(),
    classifiers=[
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
    ],
)

BUILD_DIR = 'build'
NEO4J_VERSION = '2.3.1'
DEFAULT_USERNAME = 'neo4j'
DEFAULT_PASSWORD = 'neo4j'


@task
@needs('generate_setup', 'minilib', 'setuptools.command.sdist')
def sdist():
    """Overrides sdist to make sure that our setup.py is generated."""
    pass


@task
def start_server():
    if not os.path.exists(BUILD_DIR):
        os.makedirs(BUILD_DIR)

    if not path(BUILD_DIR + '/neo4j.tar.gz').access(os.R_OK):
        print("Downloading Neo4j Server")
        urlretrieve("http://dist.neo4j.org/neo4j-community-%s-unix.tar.gz" % NEO4J_VERSION, BUILD_DIR + "/neo4j.tar.gz")

    if not path(BUILD_DIR + '/neo4j').access(os.R_OK):
        print("Unzipping Neo4j Server..")
        call(['tar', '-xf', BUILD_DIR + "/neo4j.tar.gz", '-C', BUILD_DIR])
        os.rename(BUILD_DIR + "/neo4j-community-%s" % NEO4J_VERSION, BUILD_DIR + "/neo4j")

    call([BUILD_DIR + "/neo4j/bin/neo4j", "start"])
    change_password()


@task
def stop_server():
    if path(BUILD_DIR + '/neo4j').access(os.R_OK):
        call([BUILD_DIR + "/neo4j/bin/neo4j", "stop"])


@task
def change_password():
    """
    Changes the standard password from neo4j to testing to be able to run the test suite.
    """
    basic_auth = '%s:%s' % (DEFAULT_USERNAME, DEFAULT_PASSWORD)
    try:  # Python 2
        auth = base64.encodestring(basic_auth)
    except TypeError:  # Python 3
        auth = base64.encodestring(bytes(basic_auth, 'utf-8')).decode()

    headers = {
        "Content-Type": "application/json",
        "Accept": "application/json",
        "Authorization": "Basic %s" % auth.strip()
    }
    
    response = None
    retry = 0
    while not response:  # Retry if the server is not ready yet
        sleep(1)
        con = http.HTTPConnection('localhost:7474', timeout=10)
        try:
            con.request('GET', 'http://localhost:7474/user/neo4j', headers=headers)
            response = json.loads(con.getresponse().read().decode('utf-8'))
        except ValueError:
            con.close()
        retry += 1
        if retry > 10:
            print("Could not change password for user neo4j")
            break
    if response and response.get('password_change_required', None):
        payload = json.dumps({'password': 'testing'})
        con.request('POST', 'http://localhost:7474/user/neo4j/password', payload, headers)
        print("Password changed for user neo4j")
    con.close()




import unittest

import neo4j


class TestConnection(unittest.TestCase):

    def setUp(self):
        self.conn = neo4j.connect("http://localhost:7474")
        self.conn.authorization('neo4j', 'testing')

    def test_commit(self):
        # Given
        cursor = self.conn.cursor()
        cursor.execute("CREATE (n:TestCommit {name:1337})")

        # When
        self.conn.commit()

        # Then other cursors should see it
        cursor = self.conn.cursor()
        cursor.execute("MATCH (n:TestCommit) RETURN n.name")
        self.assertEqual(cursor.fetchone(), (1337,))

        # And other connections should see it
        cursor = neo4j.connect("http://localhost:7474").cursor()
        cursor.execute("MATCH (n:TestCommit) RETURN n.name")
        self.assertEqual(cursor.fetchone(), (1337,))

    def test_read_and_commit(self):
        # Given
        cursor = self.conn.cursor()
        cursor.execute("CREATE (n:TestCommit {name:1337})")
        cursor.rowcount  # Force client to execute

        # When
        self.conn.commit()

        # Then other cursors should see it
        cursor = self.conn.cursor()
        cursor.execute("MATCH (n:TestCommit) RETURN n.name")
        self.assertEqual(cursor.fetchone(), (1337,))

        # And other connections should see it
        cursor = neo4j.connect("http://localhost:7474").cursor()
        cursor.execute("MATCH (n:TestCommit) RETURN n.name")
        self.assertEqual(cursor.fetchone(), (1337,))

    def test_rollback(self):
        # Given
        cursor = self.conn.cursor()
        cursor.execute("CREATE (n:TestRollback {name:1337})")

        # When
        self.conn.rollback()

        # Then the same cursor should not see it
        cursor.execute("MATCH (n:TestRollback) RETURN n.name")
        self.assertEqual(cursor.rowcount, 0)

        # And other connections should see it
        cursor = neo4j.connect("http://localhost:7474").cursor()
        cursor.execute("MATCH (n:TestRollback) RETURN n.name")
        self.assertEqual(cursor.rowcount, 0)


if __name__ == '__main__':
    unittest.main()



import json
import base64

from neo4j.cursor import Cursor
from neo4j.strings import ustr

try:
    from http import client as http
    from urllib.parse import urlparse
    StandardError = Exception
except ImportError:
    import httplib as http
    from urlparse import urlparse
    from exceptions import StandardError

TX_ENDPOINT = "/db/data/transaction"


def neo_code_to_error_class(code):
    if code.startswith('Neo.ClientError.Schema'):
        return Connection.IntegrityError
    elif code.startswith('Neo.ClientError'):
        return Connection.ProgrammingError
    return Connection.InternalError


def default_error_handler(connection, cursor, errorclass, errorvalue):
    if errorclass != Connection.Warning:
        raise errorclass(errorvalue)


class Connection(object):

    class Error(StandardError):
        rollback = True

    class Warning(StandardError):
        rollback = False

    class InterfaceError(Error):
        pass

    class DatabaseError(Error):
        pass

    class InternalError(DatabaseError):
        pass

    class OperationalError(DatabaseError):
        pass

    class ProgrammingError(DatabaseError):
        rollback = False

    class IntegrityError(DatabaseError):
        rollback = False

    class DataError(DatabaseError):
        pass

    class NotSupportedError(DatabaseError):
        pass

    _COMMON_HEADERS = {"Content-Type": "application/json", "Accept": "application/json", "Connection": "keep-alive"}

    def __init__(self, db_uri):
        self.errorhandler = default_error_handler
        self._host = urlparse(db_uri).netloc
        self._http = http.HTTPConnection(self._host)
        self._tx = TX_ENDPOINT
        self._messages = []
        self._cursors = set()
        self._cursor_ids = 0

    def authorization(self, username, password):
        basic_auth = '%s:%s' % (username, password)
        try:  # Python 2
            auth = base64.encodestring(basic_auth)
        except TypeError:  # Python 3
            auth = base64.encodestring(bytes(basic_auth, 'utf-8')).decode()
        self._COMMON_HEADERS.update({"Authorization": "Basic %s" % auth.strip()})

    def commit(self):
        self._messages = []
        pending = self._gather_pending()
        
        if self._tx != TX_ENDPOINT or len(pending) > 0:
            payload = None
            if len(pending) > 0:
                payload = {'statements': [{'statement': s, 'parameters': p} for (s, p) in pending]}
            response = self._deserialize(self._http_req("POST", self._tx + "/commit", payload))
            self._tx = TX_ENDPOINT
            self._handle_errors(response, self, None)

    def rollback(self):
        self._messages = []
        self._gather_pending()  # Just used to clear all pending requests
        if self._tx != TX_ENDPOINT:
            try:
                response = self._deserialize(self._http_req("DELETE", self._tx))
                self._tx = TX_ENDPOINT
                self._handle_errors(response, self, None)
            except self.OperationalError:
                # Neo.ClientError.Transaction.UnknownId
                # Unrecognized transaction id. Transaction may have timed out and been rolled back.
                pass

    def cursor(self):
        self._messages = []
        cursor = Cursor(self._next_cursor_id(), self, self._execute)
        self._cursors.add(cursor)
        return cursor

    def close(self):
        self._messages = []
        if hasattr(self, '_http') and self._http is not None:
            self._http.close()
            self._http = None

    def __del__(self):
        self.close()

    @property
    def messages(self):
        return self._messages

    def _next_cursor_id(self):
        self._cursor_ids += 1
        return self._cursor_ids

    def _gather_pending(self):
        pending = []
        for cursor in self._cursors:
            if len(cursor._pending) > 0:
                pending.extend(cursor._pending)
                cursor._pending = []
        return pending

    def _execute(self, cursor, statements):
        """"
        Executes a list of statements, returning an iterator of results sets. Each 
        statement should be a tuple of (statement, params).
        """
        payload = [{'statement': s, 'parameters': p, 'resultDataContents':['rest']} for (s, p) in statements]
        http_response = self._http_req("POST", self._tx, {'statements': payload})

        if self._tx == TX_ENDPOINT:
            self._tx = http_response.getheader('Location')

        response = self._deserialize(http_response)

        self._handle_errors(response, cursor, cursor)

        return response['results'][-1]

    def _http_req(self, method, path, payload=None, retries=2):
        serialized_payload = json.dumps(payload) if payload is not None else None

        try:
            self._http.request(method, path, serialized_payload, self._COMMON_HEADERS)
            http_response = self._http.getresponse()
        except (http.BadStatusLine, http.CannotSendRequest):
            self._http = http.HTTPConnection(self._host)
            if retries > 0:
                return self._http_req(method, path, payload, retries-1)
            self._handle_error(self, None, Connection.OperationalError, "Connection has expired.")

        if not http_response.status in [200, 201]:
            message = "Server returned unexpected response: " + ustr(http_response.status) + ustr(http_response.read())
            self._handle_error(self, None, Connection.OperationalError, message)

        return http_response

    def _handle_errors(self, response, owner, cursor):
        for error in response['errors']:
            error_class = neo_code_to_error_class(error['code'])
            error_value = ustr(error['code']) + ": " + ustr(error['message'])
            self._handle_error(owner, cursor, error_class, error_value)

    def _handle_error(self, owner, cursor, error_class, error_value):
        if error_class.rollback:
            self._tx = TX_ENDPOINT
            self._gather_pending()  # Just used to clear all pending requests
        owner._messages.append((error_class, error_value))
        owner.errorhandler(self, cursor, error_class, error_value)

    def _deserialize(self, response):
        # TODO: This is exceptionally annoying, python 3 has improved byte array handling, but that means the JSON
        # parser no longer supports deserializing these things in a streaming manner, so we have to decode the whole
        # thing first.
        return json.loads(response.read().decode('utf-8'))




# From Nigel Smalls util library
try:
    unicode
except NameError:
    # Python 3
    def ustr(s, encoding="utf-8"):
        if isinstance(s, str):
            return s
        try:
            return s.decode(encoding)
        except AttributeError:
            return str(s)

    unicode_type = str
else:
    # Python 2
    def ustr(s, encoding="utf-8"):
        if isinstance(s, str):
            return s.decode(encoding)
        else:
            return unicode(s)

    unicode_type = unicode


import unittest

import neo4j


class TestExceptions(unittest.TestCase):

    def setUp(self):
        self.conn = neo4j.connect("http://localhost:7474")
        self.conn.authorization('neo4j', 'testing')

    def test_syntax_error(self):
        # Given
        cursor = self.conn.cursor()
        
        # When
        try:
            cursor.execute("this is not valid syntax")
            cursor.rowcount  # Force client to talk to server
            raise Exception("Should not have reached here.")
        except neo4j.ProgrammingError as e:
            # Then
            self.assertEqual(str(e), "Neo.ClientError.Statement.InvalidSyntax: Invalid input \'t\': expected <init> (line 1, column 1 (offset: 0))\n\"this is not valid syntax\"\n ^")
            self.assertEqual(cursor.messages, [(neo4j.ProgrammingError, "Neo.ClientError.Statement.InvalidSyntax: Invalid input \'t\': expected <init> (line 1, column 1 (offset: 0))\n\"this is not valid syntax\"\n ^")])

    def test_cursor_clears_errors(self):
        # Given
        cursor = self.conn.cursor()
        try:
            cursor.execute("this is not valid syntax")
            cursor.rowcount  # Force client to talk to server
        except neo4j.ProgrammingError as e:
            pass
        self.conn.rollback()

        # When
        cursor.execute("CREATE (n)")

        # Then
        msg = cursor.messages
        self.assertEqual(msg, [])

if __name__ == '__main__':
    unittest.main()


import unittest

from neo4j.contextmanager import Neo4jDBConnectionManager


class TestConnectionManager(unittest.TestCase):

    def setUp(self):
        self.manager = Neo4jDBConnectionManager("http://localhost:7474", "neo4j", "testing")

    def test_commit(self):
        # Given
        with self.manager.write as cursor:
            cursor.execute("CREATE (n:TestCommit {name:1337})")

        # Other cursors should see it
        with self.manager.read as cursor:
            cursor.execute("MATCH (n:TestCommit) RETURN n.name")
            self.assertEqual(cursor.fetchone(), (1337,))

        # And other connections should see it
        with self.manager.transaction as cursor:
            cursor.execute("MATCH (n:TestCommit) RETURN n.name")
            self.assertEqual(cursor.fetchone(), (1337,))

    def test_read_and_commit(self):
        # Given
        with self.manager.write as cursor:
            cursor.execute("CREATE (n:TestCommit {name:1337})")
            cursor.rowcount  # Force client to execute

        # Then other cursors should see it
        with self.manager.read as cursor:
            cursor.execute("MATCH (n:TestCommit) RETURN n.name")
            self.assertEqual(cursor.fetchone(), (1337,))

        # And other connections should see it
        with self.manager.transaction as cursor:
            cursor.execute("MATCH (n:TestCommit) RETURN n.name")
            self.assertEqual(cursor.fetchone(), (1337,))

    def test_rollback(self):
        # Given
        with self.manager.write as cursor:
            cursor.execute("CREATE (n:TestRollback {name:1337})")
            # When
            cursor.connection.rollback()
            # Then the same cursor should not see it
            cursor.execute("MATCH (n:TestRollback) RETURN n.name")
            self.assertEqual(cursor.rowcount, 0)

        # And other connections should see it
        with self.manager.transaction as cursor:
            cursor.execute("MATCH (n:TestRollback) RETURN n.name")
            self.assertEqual(cursor.rowcount, 0)


if __name__ == '__main__':
    unittest.main()



import neo4j
import time


def main():
    conn = neo4j.connect("http://localhost:7474")
    conn.authorization('neo4j', 'testing')
    start = time.time()
    iterations = 10000
    for it in xrange(iterations):
        cursor = conn.cursor()
        for i in xrange(50):
            cursor.execute("CREATE (n:User)")
        conn.commit()
    delta = time.time() - start
    print('Tx/s: ' + str(iterations / delta))

if __name__ == "__main__":
    main()



import unittest

import neo4j
from neo4j.strings import unicode_type


class TestTypes(unittest.TestCase):

    def setUp(self):
        self.conn = neo4j.connect("http://localhost:7474")
        self.conn.authorization('neo4j', 'testing')

    def test_nodes(self):
        # Given
        cursor = self.conn.cursor()
        cursor.execute("CREATE (n:Params {name:{0}})", "Bob")

        # When
        res = list(cursor.execute("MATCH (n:Params) RETURN n"))

        # Then
        self.assertEqual(len(res), 1)
        node = res[0][0]
        self.assertEqual(node.labels, ['Params'])
        self.assertEqual(node['name'], "Bob")

    def test_rels(self):
        # Given
        cursor = self.conn.cursor()
        cursor.execute("CREATE (n:Types)-[r:KNOWS]->() SET r.name={0}", "Bob")

        # When
        res = list(cursor.execute("MATCH (:Types)-[r]-() RETURN r"))

        # Then
        self.assertEqual(len(res), 1)
        rel = res[0][0]
        self.assertEqual(rel.type, 'KNOWS')
        self.assertEqual(rel['name'], "Bob")

    def test_nested_structures(self):
        # Given
        cursor = self.conn.cursor()
        cursor.execute("CREATE (n:Types {name:{0}})-[r:KNOWS]->() SET r.name={0}", "Bob")

        # When
        res = list(cursor.execute("MATCH (n:Types)-[r]-() RETURN [{rel:r, node:n}]"))

        # Then
        self.assertEqual(len(res), 1)
        
        complicated = res[0][0]
        rel = complicated[0]['rel']
        node = complicated[0]['node']

        self.assertEqual(rel.type, 'KNOWS')
        self.assertEqual(rel['name'], "Bob")

        self.assertTrue( isinstance(rel.start_id, unicode_type) )
        self.assertTrue( isinstance(rel.end_id, unicode_type) )
        self.assertTrue( isinstance(rel.id, unicode_type) )

        self.assertEqual(node.labels, ['Types'])
        self.assertEqual(node['name'], "Bob")
        self.assertTrue( isinstance(node.id, unicode_type) )



if __name__ == '__main__':
    unittest.main()



import neo4j
from neo4j.strings import ustr


class Cursor(object):

    def __init__( self, cursorid, connection, execute_statements ):
        self.connection = connection
        self.lastrowid = None
        self.arraysize = 1
        self.errorhandler = connection.errorhandler

        self._id = cursorid

        self._pending = []
        self._execute = execute_statements
        self._rows = None
        self._rowcount = -1
        self._cursor = 0
        self._messages = []

    def execute(self, statement, *args, **kwargs):
        for i in range(len(args)):
            kwargs[i] = args[i]

        self._messages = []
        self._rows = None
        self._rowcount = 0

        self._pending.append((statement, kwargs))
        return self

    def fetchone(self):
        self._execute_pending()
        row = self._rows[self._cursor]
        self._cursor += 1
        return self._map_row(row['rest'])

    def fetchmany(self, size=None):
        self._execute_pending()
        if size is None:
            size = self.arraysize
        result = [self._map_row(r['rest']) for r in self._rows[self._cursor:self._cursor + size]]
        self._cursor += size
        return result

    def fetchall(self):
        self._execute_pending()
        result = [self._map_row(r['rest']) for r in self._rows[self._cursor:]]
        self._cursor += self.rowcount
        return result

    def __iter__(self):
        self._execute_pending()
        return self

    def __next__(self):
        try:
            return self.fetchone()
        except IndexError:
            raise StopIteration()

    def next(self):
        return self.__next__()

    def scroll(self, value, mode='relative'):
        self._execute_pending()
        if value < 0:
            raise self.connection.NotSupportedError()
        if mode == 'relative':
            self._cursor += value
        elif mode == 'absolute':
            self._cursor = value

        if self._cursor >= self.rowcount:
            self._cursor = self.rowcount
            raise IndexError()

    @property
    def description(self):
        self._execute_pending()
        return self._description

    @property
    def rowcount(self):
        self._execute_pending()
        return self._rowcount

    @property
    def messages(self):
        self._execute_pending()
        return self._messages

    def nextset(self):
        pass

    def setinputsizes(self, sizes):
        pass

    def setoutputsizes(self, size, column=None):
        pass

    def close(self):
        self._rows = None
        self._rowcount = -1
        self._messages = []
        self._description = None
        self.connection._cursors.discard(self)

    def __del__(self):
        self.close()

    def __hash__(self):
        return self._id

    def __eq__(self, other):
        return self._id == other._id

    def _map_row(self, row):
        return tuple(self._map_value(row))

    def _map_value(self, value):
        ''' Maps a raw deserialized row to proper types '''
        # TODO: Once we've gotten here, we've done the following:
        # -> Recieve the full response, copy it from network buffer it into a ByteBuffer (copy 1)
        # -> Copy all the data into a String (copy 2)
        # -> Deserialize that string (copy 3)
        # -> Map the deserialized JSON to our response format (copy 4, what we are doing in this method)
        # This should not bee needed. Technically, this mapping from transport format to python types should require exactly one copy, 
        # from the network buffer into the python VM.
        if isinstance(value, list):
            out = []
            for c in value:
                out.append(self._map_value( c ))
            return out
        elif isinstance(value, dict) and 'metadata' in value and 'labels' in value['metadata'] and 'self' in value:
            return neo4j.Node(ustr(value['metadata']['id']), value['metadata']['labels'], value['data'])
        elif isinstance(value, dict) and 'metadata' in value and 'type' in value and 'self' in value:
            return neo4j.Relationship(ustr(value['metadata']['id']), value['type'], value['start'].split('/')[-1], value['end'].split('/')[-1], value['data'])
        elif isinstance(value, dict):
            out = {}
            for k,v in value.items():
                out[k] = self._map_value( v )
            return out
        elif isinstance(value, str):
            return ustr(value)
        else:
            return value


    def _execute_pending(self):
        if len(self._pending) > 0:
            pending = self._pending

            # Clear these, in case the request fails.
            self._pending = []
            self._rows = []
            self._rowcount = 0
            self._description = []

            result = self._execute(self, pending)

            self._rows = result['data']
            self._rowcount = len(self._rows)
            self._description = [(name, neo4j.MIXED, None, None, None, None, True) for name in result['columns']]
            self._cursor = 0



from contextlib import contextmanager
from neo4j import connect


class Neo4jDBConnectionManager:

    """
    Every new connection is a transaction. To minimize new connection overhead for many reads we try to reuse a single
    connection. If this seem like a bad idea some kind of connection pool might work better.

    Neo4jDBConnectionManager.read()
    When using with Neo4jDBConnectionManager.read(): we will always rollback the transaction. All exceptions will be
    thrown.

    Neo4jDBConnectionManager.write()
    When using with Neo4jDBConnectionManager.write() we will always commit the transaction except when we see an
    exception. If we get an exception we will rollback the transaction and throw the exception.

    Neo4jDBConnectionManager.transaction()
    When we don't want to share a connection (transaction context) we can set up a new connection which will work
    just as the write context manager above but with it's own connection.

    >>> manager = Neo4jDBConnectionManager("http://localhost:7474")
    >>> with manager.write() as w:
    ...     w.execute("CREATE (TheMatrix:Movie {title:'The Matrix', tagline:'Welcome to the Real World'})")
    ...
    <neo4j.cursor.Cursor object at 0xb6fafa4c>
    >>>
    >>> with manager.read() as r:
    ...     for n in r.execute("MATCH (n:Movie) RETURN n LIMIT 1"):
    ...         print n
    "({'tagline': 'Welcome to the Real World', 'title': 'The Matrix'},)"

    Commits in batches can be achieved by:
    >>> with manager.write() as w:
    ...     w.execute("CREATE (TheMatrix:Movie {title:'The Matrix Reloaded', tagline:'Free your mind.'})")
    ...     w.connection.commit()  # The Matric Reloaded will be committed
    ...     w.execute("CREATE (TheMatrix:Movie {title:'Matrix Revolutions', tagline:'Everything that has a beginning has an end.'})")
    """

    def __init__(self, dsn, username=None, password=None):
        self.dsn = dsn
        self.connection = connect(dsn, username, password)

    @contextmanager
    def _read(self):
        cursor = self.connection.cursor()
        try:
            yield cursor
        finally:
            cursor.close()
            self.connection.rollback()
    read = property(_read)

    @contextmanager
    def _write(self):
        cursor = self.connection.cursor()
        try:
            yield cursor
        except self.connection.Error as e:
            cursor.close()
            self.connection.rollback()
            raise e
        else:
            cursor.close()
            self.connection.commit()
        finally:
            pass
    write = property(_write)

    @contextmanager
    def _transaction(self):
        connection = connect(self.dsn)
        cursor = connection.cursor()
        try:
            yield cursor
        except self.connection.Error as e:
            connection.rollback()
            raise e
        else:
            connection.commit()
        finally:
            cursor.close()
            connection.close()
    transaction = property(_transaction)



import unittest

import neo4j


class TestCursor(unittest.TestCase):

    def setUp(self):
        self.conn = neo4j.connect("http://localhost:7474")
        self.conn.authorization('neo4j', 'testing')

    def test_description(self):
        # Given
        cursor = self.conn.cursor()

        # When
        cursor.execute("CREATE (n) RETURN 1 AS hello, 'str' AS str, 3")

        # Then
        self.assertEqual(cursor.description, [
            ('hello', neo4j.MIXED, None, None, None, None, True),
            ('str',   neo4j.MIXED, None, None, None, None, True),
            ('3',     neo4j.MIXED, None, None, None, None, True)
        ])

    def test_positional_parameters(self):
        # Given
        cursor = self.conn.cursor()

        # When
        cursor.execute("CREATE (n:Params {name:{0}})", "Bob")

        # Then
        self.assertEqual(list(cursor.execute("MATCH (n:Params) RETURN n")), [({'name': 'Bob'},)])

    def test_named_parameters(self):
        # Given
        cursor = self.conn.cursor()

        # When
        cursor.execute("CREATE (n:Params {name:{name}})", name="Bob")

        # Then
        self.assertEqual(list(cursor.execute("MATCH (n:Params) RETURN n")), [({'name': 'Bob'},)])

    def test_fetch_one(self):
        # Given
        cursor = self.conn.cursor()

        # When
        cursor.execute("CREATE (n) RETURN 1,2,3")

        # Then
        self.assertEqual(cursor.rowcount, 1)
        self.assertEqual(cursor.fetchone(), (1, 2, 3))

    def test_fetch_many(self):
        # Given
        cursor = self.conn.cursor()

        # When
        cursor.execute("""FOREACH (n IN [1,2,3,4,5,6,7]| CREATE (:Test { id:n }))
                          WITH 1 AS p
                          MATCH (k:Test)
                          RETURN k.id AS id ORDER BY id""")

        # Then
        self.assertEqual(cursor.rowcount, 7)
        self.assertEqual(cursor.fetchmany(2), [(1,), (2,)])
        self.assertEqual(cursor.fetchmany(3), [(3,), (4,), (5,)])
        self.assertEqual(cursor.fetchmany(5), [(6,), (7,)])

    def test_fetch_all(self):
        # Given
        cursor = self.conn.cursor()

        # When
        cursor.execute("""FOREACH (n IN [1,2,3,4,5,6,7]| CREATE (:Test { id:n }))
                          WITH 1 AS p
                          MATCH (k:Test)
                          RETURN k.id AS id ORDER BY id""")

        # Then
        self.assertEqual(cursor.fetchone(), (1,))
        self.assertEqual(cursor.fetchall(), [(2,), (3,), (4,), (5,), (6,), (7,)])
        self.assertEqual(cursor.fetchall(), [])

    def test_iter(self):
        # Given
        cursor = self.conn.cursor()

        # When
        cursor.execute("""FOREACH (n IN [1,2,3,4,5,6,7]| CREATE (:Test { id:n }))
                          WITH 1 AS p
                          MATCH (k:Test)
                          RETURN k.id AS id ORDER BY id""")

        # Then
        cells = []
        for cell, in cursor:
            cells.append(cell)
        self.assertEqual(cells, [1, 2, 3, 4, 5, 6, 7])
        self.assertEqual(cursor.fetchall(), [])

    def test_scroll(self):
        # Given
        cursor = self.conn.cursor()
        cursor.execute("""FOREACH (n IN [1,2,3,4,5,6,7]| CREATE (:Test { id:n }))
                          WITH 1 AS p
                          MATCH (k:Test)
                          RETURN k.id AS id ORDER BY id""")
        # When
        cursor.scroll(3)

        # Then
        self.assertEqual(cursor.fetchone(), (4,))

        # And When
        cursor.scroll(5, 'absolute')

        # Then
        self.assertEqual(cursor.fetchone(), (6,))

    def test_scroll_overflow(self):
        # Given
        cursor = self.conn.cursor()
        cursor.execute("""CREATE (n) RETURN 1""")
        
        # When
        try:
            cursor.scroll(10)
            raise Exception("Should not have reached here.")
        except IndexError:
            # Then
            pass


if __name__ == '__main__':
    unittest.main()



from neo4j.connection import Connection

apilevel = '2.0'
threadsafety = 1

# This is non-standard, it uses neos built-in params. 
paramstyle = 'curly'


def connect(dsn, username=None, password=None):
    con = Connection(dsn)
    if username and password:
        con.authorization(username, password)
    return con

#
# Types
# Neo4j is a schema-optional database, which means that
# result rows can contain arbitrary types, and the database
# does not know in advance what types it will be dealing with.
# Because of this, we always describe return types as neo4j.MIXED,
# and we don't currently allow using the richer type set defined
# by the spec, since the transport format is JSON. This will change
# when the transport format changes.
#

class Node(dict):

    def __init__(self, node_id, labels, properties):
        self.id = node_id
        self.labels = labels
        for k,v in properties.items():
            self[k] = v


class Relationship(dict):

    def __init__(self, rel_id, rel_type, start_node_id, end_node_id, properties):
        self.id = rel_id
        self.type = rel_type
        self.start_id = start_node_id
        self.end_id = end_node_id
        
        for k,v in properties.items():
            self[k] = v
    

class TypeCode(object):
    
    def __init__(self, code):
        self._code = code

    def __eq__(self, other):
        return other is self

    def __unicode__(self):
        return self._code

    def __repr__(self):
        return self._code

    def __str__(self):
        return self._code


class TypeObject(object):
    
    def __init__(self, *args, **kwargs):
        raise NotSupportedError("Complex types are not yet supported.")

Date = TypeObject
Time = TypeObject
Timestamp = TypeObject
DateFromTicks = TypeObject
TimeFromTicks = TypeObject
TimestampFromTicks = TypeObject
Binary = TypeObject

STRING = TypeCode("STRING")
BINARY = TypeCode("BINARY")
NUMBER = TypeCode("NUMBER")
DATETIME = TypeCode("DATETIME")
ROWID = TypeCode("ROWID")

MIXED = TypeCode("MIXED")

#
# Exceptions 
#

Error = Connection.Error
Warning = Connection.Warning
InterfaceError = Connection.InterfaceError
DatabaseError = Connection.DatabaseError
InternalError = Connection.InternalError
OperationalError = Connection.OperationalError
ProgrammingError = Connection.ProgrammingError
IntegrityError = Connection.IntegrityError
DataError = Connection.DataError
NotSupportedError = Connection.NotSupportedError